/*
 * ******************************************************************************************
 * File		: main.c
 * Author 	: GowinSemicoductor
 * Chip		: AE350_SOC
 * Function	: Main functions
 * ******************************************************************************************
 * */

// Includes ---------------------------------------------------------------------------------
#include <stdio.h>

#include "platform.h"
#include "Driver_GPIO.h"
#include "uart.h"

#include "FreeRTOS.h"
#include "task.h"
#include "FreeRTOSConfig.h"


// Definitions ------------------------------------------------------------------------------

#define SW_STR_NAME							"FreeRTOS-AE350_SOC"
#define SW_STR_EDITION						"V10.3.1"
#define SW_STR_AUTHOR						"GowinSemiconductor"

#define LOGO_PRINT_ON

#ifdef LOGO_PRINT_ON
#include "logo.h"
#endif	//LOGO_PRINT_ON

// Task 0 thread
#define TASK0_TASK_PRIO			1
#define TASK0_STK_SIZE 			48
TaskHandle_t TASK0_Handler;
static void task0(void *pvParameters);

// Task 1 thread
#define TASK1_TASK_PRIO			2
#define TASK1_STK_SIZE 			48
TaskHandle_t TASK1_Handler;
static void task1(void *pvParameters);

extern AE350_DRIVER_GPIO 		Driver_GPIO;	// GPIO driver

#define LED0_PIN				0				// LED1
#define LED1_PIN				1				// LED2
// LED on and off flag
volatile int led0_flag = 0;
volatile int led1_flag = 0;


// Print edition
static void sw_edition_print(void)
{
	printf("\r\n");
	printf(starts_str);
	printf("Name:     "SW_STR_NAME"\r\n"
           "Edition:  "SW_STR_EDITION"\r\n"
		   "Compiled: "__DATE__", "__TIME__"\r\n"
			"Author:   "SW_STR_AUTHOR"\r\n");
	printf(starts_str);
	printf("\r\n");

#ifdef LOGO_PRINT_ON
	printf("\r\n");
	printf(starts_str);
	printf(logo_str);
	printf(starts_str);
	printf("\r\n");
#endif	//LOGO_PRINT_ON
}

static void prvSetupHardware( void )
{
	/* Ensure no interrupts execute while the scheduler is in an inconsistent
	state.  Interrupts are automatically enabled when the scheduler is
	started. */
	portDISABLE_INTERRUPTS();
}

void vAssertCalled(const char * pcFile, unsigned long ulLine)
{
	volatile unsigned long ul = 0;

	(void) pcFile;
	(void) ulLine;

	printf("ASSERT! Line %d, file %s\r\n", ( int )ulLine, pcFile);

	taskENTER_CRITICAL();
	{
		/* Set ul to a non-zero value using the debugger to step out of this
		function. */
		while (ul == 0)
		{
			portNOP();
		}
	}
	taskEXIT_CRITICAL();
}

// GPIO callback
void gpio_callback(uint32_t event)
{
	switch (event)
	{
		// Don't have GPIO pins as input interrupt event
		default:
			break;
	}
}

// Initializes GPIO
static void gpio_init(void)
{
	AE350_DRIVER_GPIO *GPIO_Dri = &Driver_GPIO;

	GPIO_Dri->Initialize(gpio_callback);	//Initializes

	GPIO_Dri->SetDir((1 << LED0_PIN) | (1 << LED1_PIN), AE350_GPIO_DIR_OUTPUT);	//Set output direction

	GPIO_Dri->PinWrite(LED0_PIN, 0);	//Set LED1 off
	GPIO_Dri->PinWrite(LED1_PIN, 0);	//Set LED2 off
}

// Task 0 thread
static void task0(void *pvParameters)
{
	AE350_DRIVER_GPIO *GPIO_Dri = &Driver_GPIO;

    while (1)
	{
		if (0 == led0_flag)
		{
			GPIO_Dri->PinWrite(LED0_PIN, 1);	//LED1 on
		}
		else
		{
			GPIO_Dri->PinWrite(LED0_PIN, 0);	//LED1 off
		}

		led0_flag = !led0_flag;

		printf("0.task0 \r\n");

		vTaskDelay(pdMS_TO_TICKS(25));
	}
}

// Task 1 thread
static void task1(void *pvParameters)
{
	AE350_DRIVER_GPIO *GPIO_Dri = &Driver_GPIO;

    while (1)
	{
		if (0 == led1_flag)
		{
			GPIO_Dri->PinWrite(LED1_PIN, 1);	//LED2 on
		}
		else
		{
			GPIO_Dri->PinWrite(LED1_PIN, 0);	//LED2 off
		}

		led1_flag = !led1_flag;

		printf("1.task1 \r\n");

		vTaskDelay(pdMS_TO_TICKS(100));
	}
}

// Main entry function
int main(void)
{
	prvSetupHardware();

	/* Initializes hardware */
	uart_init(38400);		// Initializes UART, baud rate is 38400
	gpio_init();			// Initializes GPIO
	sw_edition_print();		// Print edition

	taskENTER_CRITICAL();
	{
		// Task 0 thread
		xTaskCreate((TaskFunction_t )task0,
					(const char *   )"task0",
					(const uint16_t )TASK0_STK_SIZE,
					(void * const   )NULL,
					(UBaseType_t    )TASK0_TASK_PRIO,
					(TaskHandle_t *  )&TASK0_Handler);

		// Task 1 thread
		xTaskCreate((TaskFunction_t )task1,
					(const char *   )"task1",
					(const uint16_t )TASK1_STK_SIZE,
					(void * const   )NULL,
					(UBaseType_t    )TASK1_TASK_PRIO,
					(TaskHandle_t *  )&TASK1_Handler);
	}
    taskEXIT_CRITICAL();

	vTaskStartScheduler();

	while (1);

	return 0;
}
