/*
 * ******************************************************************************************
 * File		: trap.c
 * Author	: GowinSemicoductor
 * Chip		: AE350_SOC
 * Function	: Trap handler
 * ******************************************************************************************
 */

// No Vectored PLIC

// Includes ---------------------------------------------------------------------------------
#include <stdio.h>
#include "platform.h"


// Definitions ------------------------------------------------------------------------------

// Saved context structure
typedef struct
{
	union
	{
		struct
		{
			long x1;
			long x4;
			long x5;
			long x6;
			long x7;
			long x10;
			long x11;
			long x12;
			long x13;
			long x14;
			long x15;
			long x16;
			long x17;
			long x28;
			long x29;
			long x30;
			long x31;
		};
		long caller_regs[17];
	};
	long mepc;
	long mstatus;
	long mxstatus;
#ifdef __riscv_flen
#if __riscv_flen == 64
	long long fp_caller_regs[20];
#else
	long fp_caller_regs[20];
#endif
#endif
} SAVED_CONTEXT;

// Machine timer interrupt handler
__attribute__((weak)) void mtime_handler(void)
{
	clear_csr(NDS_MIE, MIP_MTIP);
}

// Machine software interrupt handler
__attribute__((weak)) void mswi_handler(void)
{
	clear_csr(NDS_MIE, MIP_MSIP);
}

// System call interrupt handler
__attribute__((weak)) void syscall_handler(long n, long a0, long a1, long a2, long a3)
{
	printf("syscall #%ld (a0:0x%lx,a1:0x%lx, a2:0x%lx, a3:0x%lx)\n", n, a0, a1, a2, a3);
}

// Exception interrupt handler
__attribute__((weak)) long except_handler(long cause, long epc, long *reg)
{
	/* Unhandled Trap */
	while(1);
}

// Trap interrupt handler
void trap_handler(unsigned long mcause, SAVED_CONTEXT *context)
{
	/* Do your trap handling */
	if ((mcause & MCAUSE_INT) && ((mcause & MCAUSE_CAUSE) == IRQ_M_EXT))
	{
		/* Machine-level interrupt from PLIC */
		/* External peripherals interrupt handler */
		mext_interrupt(__nds__plic_claim_interrupt());
	}
	else if ((mcause & MCAUSE_INT) && ((mcause & MCAUSE_CAUSE) == IRQ_M_TIMER))
	{
		/* Machine level PLMT interrupt */
		mtime_handler();
	}
	else if ((mcause & MCAUSE_INT) && ((mcause & MCAUSE_CAUSE) == IRQ_M_SOFT))
	{
		/* Machine level SWI interrupt */
		mswi_handler();
	}
	else if (!(mcause & MCAUSE_INT) && ((mcause & MCAUSE_CAUSE) == TRAP_M_ECALL))
	{
		/* Machine level system call interrupt */
		syscall_handler(context->x17, context->x10, context->x11, context->x12, context->x13);
		context->mepc += 4;
	}
	else
	{
		/* Unhandled Trap */
		context->mepc = except_handler(mcause, context->mepc, context->caller_regs);
	}
}
